/*****************************************************************************
* Copyright:    (c) 2024 AMFELTEC Corp.
* ----------------------------------------------------------------------------
This file is part of AMFELTEC System Logger Software Package

# AMFELTEC System Logger Software Package 
# is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# AMFELTEC System Logger Software Package
# is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License.
# If not, see <https://www.gnu.org/licenses/>

*****************************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>

#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <termios.h>
#include <sys/time.h>

#include "parser.h"
#include "parsutils.h"
#include "comport.h"
#include "utilities.h"
#include "outpututils.h"
#include "dataconv.h"
#include "configfile.h"

#define BUFFER_SIZE 128 * 1024 * 10

#if DEBUG
#define FIRST_DUMP_BYTE        '8'  
#define SECOND_DUMP_BYTE       '8'
#define FIRST_INFO_BYTE        '4'  
#define SECOND_INFO_BYTE       '4'

#define SERV_FIRST_BYTE        '6'
#define SERV_PROD_TEST_FLAG    '0'
#define SERV_WARM_RESTART_FLAG '1'
#define SERV_FIRST_START_FLAG  '2' 
#define CONF_ACEL_FLAG         '3' 
#define CONF_COM_REG_FLAG      '5'
#define CONF_INIT_DEF_FLAG     '7' 
#define LAST_CMD_BYTE          '!'

#else 

#define FIRST_DUMP_BYTE        0xC8 
#define SECOND_DUMP_BYTE       0xC8 
#define FIRST_INFO_BYTE        0xC4
#define SECOND_INFO_BYTE       0xC4

#define SERV_FIRST_BYTE        0xD1
#define SERV_PROD_TEST_FLAG    '0'
#define SERV_WARM_RESTART_FLAG '1'
#define SERV_FIRST_START_FLAG  '2' 
#define CONF_ACEL_FLAG         0xC3 
#define CONF_COM_REG_FLAG      0xC5
#define CONF_INIT_DEF_FLAG     0xC7 
#define LAST_CMD_BYTE          0xCF

#endif



char DataInComBuffer(void) {

    int comPort = GetSerialPort();
    char buff[2] = "\0";  
    int foundAt = 0;
    int foundQuestionMark = 0;
    char ret = 'X';
    int i = 0;
    int readByte;
#ifdef DEBUG    
    printf("DataInComBuffer Enter\n");
#endif
    int pcr = 0;
    do {  
        i++;
        pcr = PortComIsReady(1,0); 
        if (pcr == 0) { 
#ifdef DEBUG                
            printf("PortComIsReady() returns %d \n", pcr);
#endif            
            continue;
        }  
        readByte = read(comPort, buff, 1);

        if (buff[0] == '@') {
            foundAt = 1;
            buff[0] = '\0';
#ifdef DEBUG          
            printf(">>>>>>>>>>>>>>>> FOUND @ <<<<<<<<<<<<<<<<<");
#endif            
            break;
        } else if (buff[0] == '?') {
            foundQuestionMark = 1;
        }       
        buff[0] = '\0'; 
     } while (readByte != 0 && pcr > 0);

 
    if (foundAt == 1) {        
        ret = '@';
#ifdef DEBUG        
        printf("DataInComBuffer Exit, i = [%d], ret = %c\n", i, ret);
#endif        
        return ret; 
    } else if (foundQuestionMark == 1) {
        ret = '?';
#ifdef DEBUG        
        printf("DataInComBuffer Exit, i = [%d], ret = %c\n", i, ret);
#endif        
        return ret; 
    } else {
        ret = 'X';
    }
#ifdef DEBUG    
    printf("DataInComBuffer Exit, i = [%d], ret = %c\n", i, ret);
#endif    
    return ret; 
}


int WriteCommandToPort(unsigned char write_data[]) {

    int openedComPort = GetSerialPort();
    char bufFlag = 'x';

    if (GetComPortStatus() == -10) { 
        PortComIsReady(1,0);
    } 

    tcflush(openedComPort, TCIOFLUSH);
    write_to_serial(openedComPort, write_data, 0);
    usleep(10000);
    bufFlag = DataInComBuffer();
    if (bufFlag == '?') {
        write_to_serial(openedComPort, write_data, -1);
        usleep(10000);
        bufFlag = DataInComBuffer();
        if (bufFlag == '@') {
            return 0;
        } 
#ifdef DEBUG        
        else {
            printf("Received [%c]\n", bufFlag);
        }
#endif
    } else {
        write_to_serial(openedComPort, write_data, 1);
        usleep(10000);
            bufFlag = DataInComBuffer();
            if (bufFlag == '?') {
                write_to_serial(openedComPort, write_data, -1);
                usleep(10000);
                bufFlag = DataInComBuffer();
                if (bufFlag == '@') {
                    return 0;
                } 
#ifdef DEBUG                
                else {
                    printf("Received [%c]\n", bufFlag);
                }
#endif
            } else {
            write_to_serial(openedComPort, write_data, 2);
            usleep(10000);
            bufFlag = DataInComBuffer();
            if (bufFlag == '@') {
                return 0; 
            } else {
#ifdef DEBUG                
                printf("Received [%c], port is dead\n", bufFlag);
#endif                
                return 5;
            } 
        }
    }
    return -1; 
}

int SendBatchCommand(KeyValue* batch) {
    int retVal = 0;
    long unsigned int i = 0;
    unsigned char write_data[3];
    // a mask with bits 0 to 5 set 
    unsigned char mask = 0x3f; // 0x3f = 00111111

    write_data[2] = LAST_CMD_BYTE;
#ifdef DEBUG    
    printf("Sending Batch commands\n");
#endif    
    //for (i = 0; i < sizeof(batch)/sizeof(KeyValue); i++) {
    for (i = 0; batch[i].Key != -1; i++) {
#ifdef DEBUG        
        printf("Key(%ld) = %d, Value(%ld) = %hhu \n", i, batch[i].Key, i, batch[i].Value);
#endif        
        if (batch[i].Key != 0) {            
            if (batch[i].Key == 7) {
                batch[i].Key = CONF_INIT_DEF_FLAG;
                write_data[0] = batch[i].Key;
                write_data[1] = batch[i].Key;
            }
            else if (batch[i].Key == 5) {
                batch[i].Key = CONF_COM_REG_FLAG;
                write_data[0] = batch[i].Key;
                if ((batch[i].Value & mask) == 0) {
                    write_data[1] = batch[i].Value;
                } else {
                    write_data[1] = 0;
                }
            }
            else if (batch[i].Key == 3) {
                batch[i].Key = CONF_ACEL_FLAG;
                write_data[0] = batch[i].Key;
                write_data[1] = batch[i].Value;
            } else {     // Should never get here, unless we have more entries in config file.
                         // but in this case some additional handling will be required    
                write_data[0] = batch[i].Key;                      
                write_data[1] = batch[i].Value;
            }
            if (WriteCommandToPort(write_data) == 0) {
                retVal++;
                write_data[0] = write_data[1];
            }
        }
    }
#ifdef DEBUG    
    printf("%d commands written\n", retVal);
#endif    
    return retVal; // returns number of commands
}

int SendCommand(char* command) {

    unsigned char write_data[3];
    
    write_data[2] = LAST_CMD_BYTE;
   
    if (!strcmp(command, DUMP_CMD)) {

        write_data[0] = FIRST_DUMP_BYTE;
        write_data[1] = SECOND_DUMP_BYTE;

    } else if (!strcmp(command, INFO_CMD)) {

        write_data[0] = FIRST_INFO_BYTE;
        write_data[1] = SECOND_INFO_BYTE;

    } else if (!strcmp(command, PROD_TEST)) {

        write_data[0] = SERV_FIRST_BYTE;
        write_data[1] = SERV_PROD_TEST_FLAG;

    } else if (!strcmp(command, WARM_RESTART)) {

        write_data[0] = SERV_FIRST_BYTE;
        write_data[1] = SERV_WARM_RESTART_FLAG;

    } else if (!strcmp(command, FIRST_START)) {

        write_data[0] = SERV_FIRST_BYTE;
        write_data[1] = SERV_FIRST_START_FLAG;

    }
    return WriteCommandToPort(write_data);
}

void read_test_data_from_port(int serialPort) {
    char buffer[1024];
    char buff[2];
    int bytesRead;
    int totalBytesRead = 0;
    buffer[0] = '\0';
    do {
        PortComIsReady(2, 0);
        if (GetComPortStatus() > 0) { 
            buff[0] = '\0';
            bytesRead = read(serialPort, buff, 1);
            if (bytesRead < 0) {
                perror("Error reading from port - trace 1\n");
            }
            buffer[totalBytesRead++] = buff[0];
        } else if (GetComPortStatus() == 0) {  // 1 sec timeout 
#ifdef DEBUG        
            printf("No data received from com port within 2 seconds\n");
#endif
            ;
        } else { // reading error
            perror("Error reading from port - trace 2\n");
        }

    } while (GetComPortStatus() > 0); 
    buffer[totalBytesRead + 1] = '\0';
#ifdef DEBUG    
    printf("Finished reading, buffer size is [%ld], total bytes read [%d]\n", strlen(buffer), totalBytesRead);   
#endif    
    printf("%s\n", buffer);
}


void read_info_from_port(int serialPort) {
    char buffer[512];
    char buff[2];
    int bytesRead;
    int totalBytesRead = 0;
    buffer[0] = '\0';

    do {    

        PortComIsReady(1, 0);
        if (GetComPortStatus() > 0) { // can read here

            buff[0] = '\0';                            
            bytesRead = read(serialPort, buff, 1);
            if (bytesRead < 0) {
                perror("Error reading from port - trace 1\n");
            }
            buffer[totalBytesRead++] = buff[0]; 

        } else if (GetComPortStatus() == 0) {  // 1 sec timeout 
#ifdef DEBUG        
            printf("No data received from com port within 2 seconds\n");
#endif
            ;
        } else { // reading error
            perror("Error reading from port - trace 2\n");
        }
    } while (GetComPortStatus() > 0); 
    buffer[totalBytesRead + 1] = '\0';
#ifdef DEBUG    
    printf("Finished reading, buffer size is [%ld], total bytes read [%d]\n", strlen(buffer), totalBytesRead);   
#endif    
    printf("%s\n", buffer);
}

void read_dump_from_port(void){

    char commonArray[BUFFER_SIZE];
    char buffer[BUFFER_SIZE];
    int bytesRead = 0;
    int totalBytesRead = 0;
    int readCounter = 0;

    int openedComPort = 0;

    openedComPort = GetSerialPort();
        
    do {
        PortComIsReady(1, 0);

        if (GetComPortStatus() > 0) { // can read here
 
            buffer[0] = '\0';                
            bytesRead = read(openedComPort, buffer, sizeof(buffer) - 1);
            if (bytesRead < 0) {
                perror("Error reading from port: bytesRead < 0)\n");
            }

            totalBytesRead += bytesRead;
            readCounter++;
            printf("%c ", '#');
            if (!(readCounter % 20)) {
                printf("[%d bytes]\n", totalBytesRead);
            }           
            strncat(commonArray, buffer, strlen(buffer));
            commonArray[totalBytesRead + 1] = '\0'; 
            
        } else if (GetComPortStatus() == 0) {  // 1 sec timeout 
#ifdef DEBUG        
            printf("No data received from com port within 2 seconds\n");
#endif
            ;
        } else { // reading error
            perror("Error reading from port\n");
        }
    } while (GetComPortStatus() > 0/*  && storing == 1 */);    
#ifdef DEBUG
    printf("\nFinished reading, commonArray size is [%ld], total bytes read [%d]\n", strlen(commonArray), totalBytesRead);
#else
    printf("[%d bytes]\n", totalBytesRead);
#endif
    zero_out_file_streams();

    if (open_files() == -1) {
        perror("output files opening error, exiting" );
        return;
    }

    {
        char* copy_array = strdup(commonArray);
        if (copy_array != NULL) {
            write_to_dump_file(copy_array);
            free(copy_array);
        } 
    } 
    process_data(commonArray);
}


void process_data(char * commonArray){

    int numTokens;
    long int pressure, humidity, tempBM;
    long int x, y, z;

    char status_line[90];
    char repl_status_line[150];

    char acel_line[60];  
    char repl_acel_line[100];

    //char garb_data_line[300];
     
    status_line[0] = '\0';
    acel_line[0] = '\0';
    repl_status_line[0] = '\0';
    repl_acel_line[0] = '\0';
    //garb_data_line[0] = '\0';

    int statusIdx = 0;
    int acelIdx = 0;
    
    char *dupArray = (char *)malloc(strlen(commonArray));

    strcpy(dupArray, commonArray);
    replaceCombination(dupArray);
    removeHexCombination(dupArray);

    char **tokenArray = tokenizeString(dupArray, &numTokens);

    reverseArray(tokenArray, numTokens);

    write_status_header();
    write_accel_header();

    int idx2, idx3, idx4, idx5, idx6, idx7;

    idx2 = 2; idx3 = 3; idx4 = 4; idx5 = 5; idx6 = 6; idx7 = 7;


    for (int t = 0; t < numTokens; t++) {
        int stBlock = status_block(tokenArray, t, numTokens);
        int acBlock = acel_block(tokenArray, t, numTokens);
        
        if (stBlock != 0) {
            if (stBlock == STATUS_DATA_SIZE + 3) {
#ifdef DEBUG                
                printf("Inside status block\n");
                for (int i = 0; i < STATUS_DATA_SIZE + 4; i++){
                    printf("[%s] ", tokenArray[t + i]);
                }
                printf("\n");
#endif                

                pressure = getPressure(strtol(tokenArray[t+idx4], NULL, 16),strtol(tokenArray[t+idx5], NULL, 16));
                humidity = getHumidity(strtol(tokenArray[t+idx2], NULL, 16), strtol(tokenArray[t+idx3], NULL, 16));
                tempBM = getBmTemperature(strtol(tokenArray[t+idx6], NULL, 16),strtol(tokenArray[t+idx7], NULL, 16));
            
                sprintf(status_line, "|  %6d | %12ld | %10ld | %10ld |\n", 
                                    statusIdx, tempBM, humidity, pressure);
                sprintf(repl_status_line, "\nAMF(S): %d; tp=%ld,hm=%ld,pr=%ld\n", 
                                    statusIdx, tempBM, humidity, pressure);
            

                status_line[strlen(status_line)] = '\0';
                repl_status_line[strlen(repl_status_line)] = '\0';

                tokenArray[t] = strdup(repl_status_line);
            
                add_line_to_dump_file(tokenArray[t]);                   
                add_line_to_status_file(status_line);

                status_line[0] = '\0';  
                repl_status_line[0] = '\0';
                statusIdx++;    
                t += STATUS_DATA_SIZE + 3;
            } 
#ifdef DEBUG            
            else {
                printf("Out of size status block\n");
                for (int i = 0; i <= stBlock; i++){
                    printf("[%s] ", tokenArray[t + i]);
                }
                printf("\n");
            }
#endif
        } else if (acBlock != 0) {
            if (acBlock == ACEL_DATA_SIZE + 2) {
#ifdef DEBUG                
                printf("Inside acel block\n");
                for (int i = 0; i < ACEL_DATA_SIZE + 3; i++){
                    printf("[%s] ", tokenArray[t + i]);
                }
                printf("\n");
#endif
                x = strtol(tokenArray[t+2], NULL, 16);
                y = strtol(tokenArray[t+3], NULL, 16);
                z = strtol(tokenArray[t+4], NULL, 16);

                sprintf(acel_line, "|  %6d | %10ld | %10ld | %10ld |\n", acelIdx, x, y, z);
                sprintf(repl_acel_line, "\nAMF(A): %d; %ld, %ld, %ld\n", acelIdx, x, y, z);

                tokenArray[t] = strdup(repl_acel_line);
                add_line_to_dump_file(tokenArray[t]);

                add_line_to_acel_file(acel_line);
                acel_line[0] = '\0';  
                repl_acel_line[0] = '\0'; 

                acelIdx++;    
                t += ACEL_DATA_SIZE + 2;

            } 
#ifdef DEBUG            
            else {
                printf("Out of size acel block\n");
                for (int i = 0; i <= acBlock; i++){
                    printf("[%s] ", tokenArray[t + i]);
                }
                printf("\n");                
            } 
#endif
        } else { 
            if (acBlock == 0 &&  stBlock == 0) {
                char str[5];
                memset(str, 0, 5);                
                if (containsOnlyPrintableAscii(tokenArray[t]) == 1 ) {

                    str[0] = (char)strtol(tokenArray[t], NULL, 16);
                    add_line_to_dump_file(str);
                } else {  
                    sprintf(str, "[%s] ", tokenArray[t]);
                    add_line_to_dump_file(str);
                }
            }
        }
    }
    for (int l = 0; l < numTokens; l++) {
        if (tokenArray[l] != NULL) {
            free(tokenArray[l]);
            tokenArray[l] = NULL;
        }
    }
    if (dupArray != NULL) {
        free(dupArray);
        dupArray = NULL;
    }
    close_files(); 
    post_process_finarray();    

}