/*****************************************************************************
* Copyright:    (c) 2024 AMFELTEC Corp.
* ----------------------------------------------------------------------------
This file is part of AMFELTEC System Logger Software Package

# AMFELTEC System Logger Software Package 
# is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# AMFELTEC System Logger Software Package
# is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License.
# If not, see <https://www.gnu.org/licenses/>

*****************************************************************************/

#include <string.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <regex.h>

#include "parsutils.h"
#include "outpututils.h"

#define CONVERT_TIMESTAMP   // remove underscore if want to convert timestamp

// returns idx of first entry of start flags pair, otherwise -1
int status_block_start(char* array[], int idx, int maxIdx) {
    int startIdxS = -1;

    if (!strncmp(STATUS_START_FLAG_1, array[idx], strlen(STATUS_START_FLAG_1)) && 
        !strncmp(STATUS_START_FLAG_2, array[idx + 1], strlen(STATUS_START_FLAG_2)) &&
        idx + 9 < maxIdx
    ) {       
        startIdxS = idx;
    }

    return startIdxS;
}

// returns idx of first entry of end flags pair, otherwise -1
int status_block_end(char* array[], int idx, int maxIdx) {
    int endIdxS = -1;

    if (idx + 9 < maxIdx && 
        !strncmp(STATUS_END_FLAG_1, array[idx + 8], strlen(STATUS_END_FLAG_1)) && 
        !strncmp(STATUS_END_FLAG_2, array[idx + 9], strlen(STATUS_END_FLAG_2))
    ) {       
        endIdxS = idx + 9;
    }

    return endIdxS;
}

// returns idx of first entry of start flags pair, otherwise -1
int acel_block_start(char* array[], int idx, int maxIdx) {
    int startIdxA = -1;

    if (idx + 5 < maxIdx &&
        !strncmp(ACEL_START_FLAG_1, array[idx], strlen(ACEL_START_FLAG_1)) && 
        !strncmp(ACEL_START_FLAG_2, array[idx + 1], strlen(ACEL_START_FLAG_2))
    ) {   
        startIdxA = idx;
    }

    return startIdxA;
}

// returns idx of end flag, otherwise -1
int acel_block_end(char* array[], int idx, int maxIdx) {
    int endIdxA = -1;

    if (idx + 5 < maxIdx &&
        !strncmp(ACEL_END_FLAG, array[idx + 5], strlen(ACEL_END_FLAG)) 
    ) {
        endIdxA = idx + 5;
    }
    
    return endIdxA;
}

// extracts status block
// returns block size
int status_block(char* array[], int idx, int maxIdx) {
    
    int retVal = 0;

    int startIdx = status_block_start(array, idx, maxIdx);
    int endIdx = status_block_end(array, idx, maxIdx);
    if (startIdx != -1 && endIdx != -1) {
        retVal = endIdx - startIdx; 
    }
    return retVal;
}

// extracts accelleration block
// returns block size
int acel_block(char* array[], int idx, int maxIdx) {
 
    int retVal = 0;

    int startIdx = acel_block_start(array, idx, maxIdx);
    int endIdx = acel_block_end(array, idx, maxIdx);
    if (startIdx != -1 && endIdx != -1) {
        retVal = endIdx - startIdx; 
    }
    return retVal;
}

int containsOnlyPrintableAscii(const char *str) {

    int code = strtol(str, NULL, 16);
    if ((code < 32 || code > 126) && (code != 10 && code != 13 && code != 11) ) {   
        return 0;  // Not a printable ASCII character
    }
    return 1;  // All characters are printable ASCII
}

#define  MAX_LINE_LENGTH 2048

// Function to get the last boot time from /proc/uptime
time_t get_last_boot_time() {
    FILE *fp;
    double uptime_seconds;
    time_t current_time;

    fp = fopen("/proc/uptime", "r");
    if (fp == NULL) {
        perror("Error opening /proc/uptime");
        exit(1);
    }
    // Read the uptime (first value) from the file
    fscanf(fp, "%lf", &uptime_seconds);
    // Close the file
    fclose(fp);
    // Get the current time
    current_time = time(NULL);
    // Calculate the last boot time
    return current_time - (time_t)uptime_seconds;
}

// Function to calculate date timestamp seconds behind current 
void calculate_date(char *line) {
    char *timestamp_start = strchr(line, '[');
    char *timestamp_end = strchr(line, ']');
    char *pattern = "\\[\\s*[0-9]+\\.[0-9]+\\]";
    regex_t regex;

    if (regcomp(&regex, pattern, REG_EXTENDED) != 0) {
        fprintf(stderr, "Failed to compile regular expression\n");
        return;
    }
    char *modified_line = line;
 
    if (timestamp_start != NULL && timestamp_end != NULL ) {
        unsigned long sec;
        unsigned long msec;
        char *endptr;
        
        char orig_timestamp[20];
        char formatted_time[20];
        char rest_of_line[MAX_LINE_LENGTH];

        // Extract seconds and milliseconds from the timestamp
        sec = strtoul(timestamp_start + 1, &endptr, 10);
        msec = strtoul(endptr + 1, NULL, 10);
        sprintf(orig_timestamp, "[%ld.%ld]", sec, msec);

        if (regexec(&regex, orig_timestamp, 0, NULL, 0) == 0) {
            
            time_t current_time;
            time(&current_time);

            time_t target_time = current_time - sec;
            strftime(formatted_time, sizeof(formatted_time), "%Y-%m-%d %H:%M:%S", localtime(&target_time));
            strncpy(rest_of_line, timestamp_end + 1, sizeof(rest_of_line));
            rest_of_line[sizeof(rest_of_line) - 1] = '\0';

            char *seconds_end = strchr(formatted_time, ':');
            if (seconds_end != NULL) {
                modified_line = line + (timestamp_start - line);
                int insert_pos = seconds_end - formatted_time + 6; // After seconds and the dot
                sprintf(modified_line, "[%.*s ]%s", insert_pos, formatted_time, rest_of_line);
            }
        }
    }
}

void convert_timestamp(char *line, time_t last_boot_time) {
    char *timestamp_start = strchr(line, '[');
    char *timestamp_end = strchr(line, ']');
    char *pattern = "\\[\\s*[0-9]+\\.[0-9]+\\]";
    regex_t regex;
    
    if (regcomp(&regex, pattern, REG_EXTENDED) != 0) {
        fprintf(stderr, "Failed to compile regular expression\n");
        return;
    }
    char *modified_line = line;
 
    if (timestamp_start != NULL && timestamp_end != NULL ) {
        unsigned long sec;
        unsigned long msec;
        char *endptr;
        struct tm *timeinfo;
        char orig_timestamp[20];
        char formatted_time[20];
        char rest_of_line[MAX_LINE_LENGTH];

        // Extract seconds and milliseconds from the timestamp
        sec = strtoul(timestamp_start + 1, &endptr, 10);
        msec = strtoul(endptr + 1, NULL, 10);
        sprintf(orig_timestamp, "[%ld.%ld]", sec, msec);
        if (regexec(&regex, orig_timestamp, 0, NULL, 0) == 0) {
        // Calculate the absolute timestamp
            time_t absolute_time = last_boot_time + sec;

            // Convert the absolute timestamp to struct tm
            timeinfo = localtime(&absolute_time);

            // Format the time as "Mon Day HH:MM:SS"
            strftime(formatted_time, sizeof(formatted_time), "%b %d %H:%M:%S", timeinfo);

            // Extract the rest of the line after the timestamp
            strncpy(rest_of_line, timestamp_end + 1, sizeof(rest_of_line));
            rest_of_line[sizeof(rest_of_line) - 1] = '\0';

            // Find the position to insert milliseconds
            char *seconds_end = strchr(formatted_time, ':');
            if (seconds_end != NULL) {
                modified_line = line + (timestamp_start - line);
                int insert_pos = seconds_end - formatted_time + 6; // After seconds and the dot
                sprintf(modified_line, "[%.*s.%06lu]%s", insert_pos, formatted_time, msec, rest_of_line);
            }
        }
    }
}

// Function to remove color changing combinations
void remove_color(char *str) {
    // Define the target substrings to remove
    const char *start_substr = "[0x1b] [0;";
    const char *end_substr = "[0x1b] [0m";

    // Initialize pointers for iterating through the string
    char *read_ptr = str;
    char *write_ptr = str;

    // Iterate through the string
    while (*read_ptr != '\0') {
        // Check if the start substring is found
        if (strncmp(read_ptr, start_substr, strlen(start_substr)) == 0) {
            // Skip characters until 'm' is found or the end of the string is reached
            while (*read_ptr != 'm' && *read_ptr != '\0') {
                ++read_ptr;
            }
            // If 'm' is found, skip it as well
            if (*read_ptr == 'm') {
                ++read_ptr;
            }
        }
        // Check if the end substring is found
        else if (strncmp(read_ptr, end_substr, strlen(end_substr)) == 0) {
            // Skip characters until 'm' is found or the end of the string is reached
            while (*read_ptr != 'm' && *read_ptr != '\0') {
                ++read_ptr;
            }
            // If 'm' is found, skip it as well
            if (*read_ptr == 'm') {
                ++read_ptr;
            }
        }
        // If neither start nor end substring is found, copy the character to the final string
        else {
            *write_ptr++ = *read_ptr++;
        }
    }
    // Terminate the final string
    *write_ptr = '\0';
}

int post_process_finarray(void) {
    FILE *input_fp, *output_fp;
    char line[MAX_LINE_LENGTH];
   
   // time_t last_boot_time = get_last_boot_time();

    input_fp = fopen(FINAL_DUMP_FILE, "r");
    if (input_fp == NULL) {
        perror("Error opening input file");
        return 1;
    }

    output_fp = fopen(FINAL_LOG, "w");
    if (output_fp == NULL) {
        perror("Error opening output file");
        fclose(input_fp);
        return 1;
    }
    printf("Parsing final output\n");
    while (fgets(line, sizeof(line), input_fp) != NULL) {
        // Check if the line contains a timestamp string
        remove_color(line);

        //convert_timestamp(line, last_boot_time);
        //calculate_date(line);

        // Write the line to the output file
        if (line[0] != '\n' && line[0] != '\r') {
            fputs(line, output_fp);
        }
    }
    fclose(input_fp);
    fclose(output_fp);
    remove(FINAL_DUMP_FILE);
    return 0;
}