/*****************************************************************************
* Copyright:    (c) 2024 AMFELTEC Corp.
* ----------------------------------------------------------------------------
This file is part of AMFELTEC System Watchdog Software Package

# AMFELTEC System Watchdog Software Package 
# is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# AMFELTEC System Watchdog Software Package
# is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License.
# If not, see <https://www.gnu.org/licenses/>

*****************************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>

#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <termios.h>
#include <sys/time.h>


#include "comport.h"
#include "command.h"




#ifdef DEBUG

    #define CMD_WD_INTVL_SEC    0x30 // '0' T1
    #define CMD_RST_INTVL       0x31 // '1' T2
    #define CMD_PWR_OFF_INTVL   0x32 // '2' T3
    #define CMD_PWR_WAIT_INTVL  0x33 // '3' T4
    #define CMD_PWR_ON_INTVL    0x35 // '5' T5
    #define CMD_GET_INFO        0x34 // '4'
    #define CMD_WD_CNTL         0x36 // '6'
    #define CMD_WD_TST          0x61 // 'a'

    #define CMD_FLG_END         0x21 // '!'

#else

    #define CMD_WD_INTVL_SEC    0xC8 // '0'  T1
    #define CMD_RST_INTVL       0xC9 // '1'  T2
    #define CMD_PWR_OFF_INTVL   0xCA // '2'  T3
    #define CMD_PWR_WAIT_INTVL  0xCB // '3'  T4
    #define CMD_PWR_ON_INTVL    0xCC // '5'  T5
    #define CMD_GET_INFO        0xCD // '4'
    #define CMD_WD_CNTL         0xCE // '6'
    #define CMD_WD_TST          0xD1 // 'a'

    #define CMD_FLG_END         0xCF // '!'

#endif

#define DUMMY_BYTE              '_'  

const char timer_commands[]  = {CMD_WD_INTVL_SEC,
                                CMD_RST_INTVL,
                                CMD_PWR_OFF_INTVL,
                                CMD_PWR_WAIT_INTVL,
                                CMD_PWR_ON_INTVL};

int SendTimerCommand(char timers[]){

    char write_data[3];
    write_data[2] = CMD_FLG_END;
    int intermRet = 0;
    int i;

    for (i = 0; i < NUMBER_OF_TIMERS; i++) {
        write_data[0] = timer_commands[i];
        write_data[1] = timers[i];
        intermRet = WriteCommandToPort(write_data);
        if (intermRet != 0) {
           printf("Problem with writing timer %d - %c\n", i, write_data[1]); 
           return intermRet;
        }
    }
    return 0;
}

int setBit(int n, int k) 
{ 
    return (n | (1 << (k - 1))); 
} 

int clearBit(int n, int k) 
{ 
    return (n & (~(1 << (k - 1)))); 
} 

// receives char '0' or '1' only
// bit 0 = 1 always
// bit 1 = 1 power toggle
//         0
// bit 2 = 1 enable reset
//         0
// bit 3 = 0 - normal, 1 - unlimited cycle - not recommended
  
int SendControlCommand(char r, char p, char u) {
    char write_data[3];

    write_data[0] = CMD_WD_CNTL;
    write_data[1] = 1;
    write_data[2] = CMD_FLG_END;
    
    if (r == '1') {
        write_data[1] = setBit(write_data[1], 2);
    } else if (r == '0') {
        write_data[1] = clearBit(write_data[1], 2);
    } else {
        printf("Wrong input parameter, r should be 0 or 1\n");
        return -1;
    }

    if (p == '1') {
        write_data[1] = setBit(write_data[1], 3);
    } else if (p == '0') {
        write_data[1] = clearBit(write_data[1], 3);
    }else {
        printf("Wrong input parameter, p should be 0 or 1\n");
        return -1;
    }

    if (u == '1') {
        write_data[1] = setBit(write_data[1], 4);
    } else if (u == '0') {
        write_data[1] = clearBit(write_data[1], 4);
    } else {
        printf("Wrong input parameter, u should be 0 or 1\n");
        return -1;
    }
    printf("Byte 1 is %d\n", write_data[1]);
    return WriteCommandToPort(write_data);

}

int SendCommand(char* command) {

    char write_data[3];

    write_data[2] = CMD_FLG_END;

    if (!strcmp(command, INFO_WD_CMD)) {

        write_data[0] = CMD_GET_INFO;
        write_data[1] = CMD_GET_INFO;

    } else if (!strcmp(command, START_WD_CMD)) {
    // !!! 7 - def- no params
        write_data[0] = CMD_WD_CNTL;
        write_data[1] = 7; // LG

    } else if (!strcmp(command, STOP_WD_CMD)) {

        write_data[0] = CMD_WD_CNTL;
        write_data[1] = 0;
    }
    return WriteCommandToPort(write_data);
}

void SendRefresh(int sec) {
    char dummyByte = DUMMY_BYTE;

    if (sec == 0) {
        write_to_serial(GetSerialPort(), &dummyByte, -1);
    } else {
        while (1) {
           write_to_serial(GetSerialPort(), &dummyByte, -1);
           sleep(sec); 
        }
    }
}